'''

  NumbaCoefficient

   utility to use NumbaCoefficient more easily

'''
import traceback

from numpy.linalg import inv, det
from numpy import conj as npconj
from numpy import dot as npdot
from numpy import array, zeros, iscomplexobj
from petram.mfem_config import use_parallel

if use_parallel:
    import mfem.par as mfem
    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank

else:
    import mfem.ser as mfem
    myid = 0

from petram.helper.variables import (Variable,
                                     NativeCoefficientGenBase)


import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('NumbaCoefficient')


class NumbaCoefficient():
    def __init__(self, coeff):
        self.is_complex = coeff.IsOutComplex()
        self.mfem_numba_coeff = coeff

        if self.is_complex:
            self.real = coeff.real
            self.imag = coeff.imag
        else:
            self.real = coeff
            self.imag = None

        if self.ndim == 1:
            self._V = mfem.Vector(self.shape[0])
        if self.ndim == 2:
            self._K = mfem.DenseMatrix(self.shape[0], self.shape[1])

    @property
    def complex(self):
        return self.is_complex

    def get_real_coefficient(self):
        return self.real

    def get_imag_coefficient(self):
        return self.imag

    def get_realimag_coefficient(self, real):
        if real:
            return self.get_real_coefficient()
        else:
            return self.get_imag_coefficient()

    # @property
    # def sdim(self):
    #    return self.mfem_numba_coeff.SpaceDimension()

    @property
    def ndim(self):
        return self.mfem_numba_coeff.GetNDim()

    @property
    def shape(self):
        if self.ndim == 0:
            return tuple()
        if self.ndim == 1:
            return (self.mfem_numba_coeff.GetVDim(),)
        if self.ndim == 2:
            return (self.mfem_numba_coeff.GetWidth(),
                    self.mfem_numba_coeff.GetHeight(),)

        assert False, "unsupported dim"
        return None

    @property
    def width(self):
        return self.mfem_numba_coeff.GetWidth()

    @property
    def height(self):
        return self.mfem_numba_coeff.GetHeight()

    @property
    def vdim(self):
        return self.mfem_numba_coeff.GetVDim()

    @property
    def kind(self):
        if self.ndim == 0:
            return "scalar"
        if self.ndim == 1:
            return "vector"
        if self.ndim == 2:
            return "matrix"
        assert False, "unsupported dim"
        return None

    def eval(self, T, ip):
        if self.ndim == 0:
            if self.real is not None:
                ret = self.coeff1.Eval(T, ip)
            else:
                ret = 0
            if self.imag is not None:
                ret = ret + 1j * self.imag.Eval(T, ip)
            return ret
        if self.ndim == 1:
            if self.real is not None:
                self.real.Eval(self._V, T, ip)
            else:
                self._V.Assign(0.0)
            vec = self._V.GetDataArray().copy()
            if self.imag is not None:
                self.imag.Eval(self._V, T, ip)
                vec = vec + 1j * self._V.GetDataArray()
            return vec
        if self.ndim == 2:
            if self.real is not None:
                self.real.Eval(self._K, T, ip)
            else:
                self._K.Assign(0.0)
            mat = self._K.GetDataArray().copy()
            if self.imag is not None:
                self.imag.Eval(self._K, T, ip)
                mat = mat + 1j * self._K.GetDataArray()
            return mat
        return None

    def is_matrix(self):
        return self.mfem_numba_coeff.GetNdim() == 2

    def is_vector(self):
        return self.mfem_numba_coeff.GetNdim() == 1

    def __add__(self, other):
        '''
        ruturn sum coefficient
        '''
        from petram.phys.phys_model import (PhysConstant,
                                            PhysVectorConstant,
                                            PhysMatrixConstant,)
        from petram.phys.pycomplex_coefficient import (PyComplexConstant,
                                                       PyComplexVectorConstant,
                                                       PyComplexMatrixConstant,)
        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        if not isinstance(other, NumbaCoefficient):
            if isinstance(other, (PhysConstant,
                                  PhysVectorConstant,
                                  PhysMatrixConstant,
                                  PyComplexConstant,
                                  PyComplexVectorConstant,
                                  PyComplexMatrixConstant,)):
                params = {"value": other.value}
                dep = (self.mfem_numba_coeff, )
                func = '\n'.join(['def f(ptx, coeff1):',
                                  '    return coeff1 + value'])

            else:
                return NotImplemented
        else:
            assert self.shape == other.shape, "ndim must match to perform sum operation"
            dep = (self.mfem_numba_coeff, other.mfem_numba_coeff)
            params = None
            func = '\n'.join(['def f(ptx, coeff1, coeff2):',
                              '    return coeff1 + coeff2'])

        l = {}
        exec(func, globals(), l)
        if self.ndim == 0:
            coeff = mfem.jit.scalar(complex=self.complex,
                                    dependency=dep,
                                    params=params,
                                    debug=numba_debug)(l["f"])
        elif self.ndim == 1:
            coeff = mfem.jit.vector(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    params=params,
                                    shape=self.shape)(l["f"])

        elif self.ndim == 2:
            coeff = mfem.jit.matrix(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    params=params,
                                    shape=self.shape)(l["f"])

        else:
            assert False, "unsupported dim: dim=" + str(self.ndim)

        return NumbaCoefficient(coeff)

    def __sub__(self, other):
        '''
        ruturn sum coefficient
        '''
        from petram.phys.phys_model import (PhysConstant,
                                            PhysVectorConstant,
                                            PhysMatrixConstant,)
        from petram.phys.pycomplex_coefficient import (PyComplexConstant,
                                                       PyComplexVectorConstant,
                                                       PyComplexMatrixConstant,)
        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        if not isinstance(other, NumbaCoefficient):
            if isinstance(other, (PhysConstant,
                                  PhysVectorConstant,
                                  PhysMatrixConstant,
                                  PyComplexConstant,
                                  PyComplexVectorConstant,
                                  PyComplexMatrixConstant,)):
                params = {"value": other.value}
                dep = (self.mfem_numba_coeff, )
                func = '\n'.join(['def f(ptx, coeff1):',
                                  '    return coeff1 - value'])

            else:
                return NotImplemented
        else:
            assert self.shape == other.shape, "ndim must match to perform sum operation"
            dep = (self.mfem_numba_coeff, other.mfem_numba_coeff)
            params = None
            func = '\n'.join(['def f(ptx, coeff1, coeff2):',
                              '    return coeff1 - coeff2'])

        l = {}
        exec(func, globals(), l)
        if self.ndim == 0:
            coeff = mfem.jit.scalar(complex=self.complex,
                                    dependency=dep,
                                    params=params,
                                    debug=numba_debug)(l["f"])
        elif self.ndim == 1:
            coeff = mfem.jit.vector(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    params=params,
                                    shape=self.shape)(l["f"])

        elif self.ndim == 2:
            coeff = mfem.jit.matrix(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    params=params,
                                    shape=self.shape)(l["f"])

        else:
            assert False, "unsupported dim: dim=" + str(self.ndim)

        return NumbaCoefficient(coeff)

    def __mul__(self, other):
        from petram.phys.phys_model import (PhysConstant,
                                            PhysVectorConstant,
                                            PhysMatrixConstant,)
        from petram.phys.pycomplex_coefficient import (PyComplexConstant,
                                                       PyComplexVectorConstant,
                                                       PyComplexMatrixConstant,)

        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        dep = (self.mfem_numba_coeff, )

        if not isinstance(other, NumbaCoefficient):
            if isinstance(other, (PhysConstant,
                                  PhysVectorConstant,
                                  PhysMatrixConstant,
                                  PyComplexConstant,
                                  PyComplexVectorConstant,
                                  PyComplexMatrixConstant,)):
                params = {"scale": other.value}
            else:
                params = {"scale": other}  # this assuems other is a number

            func = '\n'.join(['def f(ptx, val):',
                              '    return val*scale'])
            is_other_complex = iscomplexobj(params["scale"])

        else:
            assert len(other.shape) == 0, "scale must be scalar (__mul__)"
            dep = (self.mfem_numba_coeff, other.mfem_numba_coeff)
            params = None
            func = '\n'.join(['def f(ptx, val, scale):',
                              '    return val*scale'])
            is_other_complex = other.complex

        l = {}
        if numba_debug:
            print("(DEBUG) numba function\n", func)
        exec(func, globals(), l)

        do_complex = self.complex or is_other_complex

        if self.ndim == 0:
            coeff = mfem.jit.scalar(complex=do_complex,
                                    dependency=dep,
                                    interface="simple",
                                    params=params,
                                    debug=numba_debug)(l["f"])
        elif self.ndim == 1:
            coeff = mfem.jit.vector(complex=do_complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    params=params,
                                    shape=self.shape)(l["f"])

        elif self.ndim == 2:
            coeff = mfem.jit.matrix(complex=do_complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    params=params,
                                    shape=self.shape)(l["f"])

        else:
            assert False, "unsupported dim: dim=" + str(self.ndim)

        return NumbaCoefficient(coeff)

    def __rmul__(self, scale):
        return self.__mul__(scale)

    def __div__(self, scale):
        return self.__truediv__(scale)

    def __truediv__(self, other):
        from petram.phys.phys_model import (PhysConstant,
                                            PhysVectorConstant,
                                            PhysMatrixConstant,)
        from petram.phys.pycomplex_coefficient import (PyComplexConstant,
                                                       PyComplexVectorConstant,
                                                       PyComplexMatrixConstant,)

        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        dep = (self.mfem_numba_coeff, )

        if not isinstance(other, NumbaCoefficient):
            if isinstance(ohter, (PhysConstant,
                                  PhysVectorConstant,
                                  PhysMatrixConstant,
                                  PyComplexConstant,
                                  PyComplexVectorConstant,
                                  PyComplexMatrixConstant,)):
                params = {"scale": other.value}
            else:
                params = {"scale": other}  # this assuems other is a number

            func = '\n'.join(['def f(ptx, val):',
                              '    return val/scale'])
            is_other_complex = iscomplexobj(params["scale"])

        else:
            assert len(other.shape) == 0, "scale must be scalar (__truediv__)"
            dep = (self.mfem_numba_coeff, other.mfem_numba_coeff)

            params = None
            func = '\n'.join(['def f(ptx, val, scale):',
                              '    return val/scale'])
            is_other_complex = other.complex

        l = {}
        if numba_debug:
            print("(DEBUG) numba function\n", func)
        exec(func, globals(), l)

        do_complex = self.complex or is_other_complex

        if self.ndim == 0:
            coeff = mfem.jit.scalar(complex=do_complex,
                                    dependency=dep,
                                    interface="simple",
                                    params=params,
                                    debug=numba_debug)(l["f"])
        elif self.ndim == 1:
            coeff = mfem.jit.vector(complex=do_complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    params=params,
                                    shape=self.shape)(l["f"])

        elif self.ndim == 2:
            coeff = mfem.jit.matrix(complex=do_complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    params=params,
                                    shape=self.shape)(l["f"])

        else:
            assert False, "unsupported dim: dim=" + str(self.ndim)

        return NumbaCoefficient(coeff)

    def __rdiv__(self, scale):
        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        func = '\n'.join(['def f(ptx, val):',
                          '    return scale/val'])

        l = {}
        if numba_debug:
            print("(DEBUG) numba function\n", func)
        exec(func, globals(), l)

        dep = (self.mfem_numba_coeff, )
        params = {'scale': scale}

        if self.ndim == 0:
            coeff = mfem.jit.scalar(complex=self.complex,
                                    dependency=dep,
                                    interface="simple",
                                    params=params,
                                    debug=numba_debug)(l["f"])
        elif self.ndim == 1:
            coeff = mfem.jit.vector(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    params=params,
                                    shape=self.shape)(l["f"])

        elif self.ndim == 2:
            coeff = mfem.jit.matrix(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    params=params,
                                    shape=self.shape)(l["f"])

        else:
            assert False, "unsupported dim: dim=" + str(self.ndim)

        return NumbaCoefficient(coeff)

    def __rtruediv__(self, scale):
        return self.__rdiv__(scale)

    def __pos__(self):
        raise NotImplementedError

    def __neg__(self):
        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        func = '\n'.join(['def f(ptx, val):',
                          '    return -val'])

        l = {}
        if numba_debug:
            print("(DEBUG) numba function\n", func)
        exec(func, globals(), l)

        dep = (self.mfem_numba_coeff, )

        if self.ndim == 0:
            coeff = mfem.jit.scalar(complex=self.complex,
                                    dependency=dep,
                                    interface="simple",
                                    debug=numba_debug)(l["f"])
        elif self.ndim == 1:
            coeff = mfem.jit.vector(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    shape=self.shape)(l["f"])

        elif self.ndim == 2:
            coeff = mfem.jit.matrix(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    shape=self.shape)(l["f"])

        else:
            assert False, "unsupported dim: dim=" + str(self.ndim)

        return NumbaCoefficient(coeff)

    def conj(self):
        '''
        generate conjugate
        '''
        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        func = '\n'.join(['def f(ptx, val):',
                          '    return npconj(val)'])

        l = {}
        if numba_debug:
            print("(DEBUG) numba function\n", func)
        exec(func, globals(), l)

        dep = (self.mfem_numba_coeff, )

        if self.ndim == 0:
            coeff = mfem.jit.scalar(complex=self.complex,
                                    dependency=dep,
                                    interface="simple",
                                    debug=numba_debug)(l["f"])
        elif self.ndim == 1:
            coeff = mfem.jit.vector(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    shape=self.shape)(l["f"])

        elif self.ndim == 2:
            coeff = mfem.jit.matrix(complex=self.complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    shape=self.shape)(l["f"])

        else:
            assert False, "unsupported dim: dim=" + str(self.ndim)

        return NumbaCoefficient(coeff)

    def dot(self, other):
        '''
        dot product matrix*matrix etc
        '''
        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        assert len(self.shape) > 0, ".dot is defined only for vector/matrix"

        if not isinstance(other, NumbaCoefficient):
            if isinstance(ohter, (PhysVectorConstant,
                                  PhysMatrixConstant,
                                  PyComplexVectorConstant,
                                  PyComplexMatrixConstant,)):
                params = {"scale": other.value}
            else:
                params = {"scale": other}  # this assuems other is a number

            func = '\n'.join(['def f(ptx, val):',
                              '    return npdot(val, scale)'])
            is_other_complex = iscomplexobj(params["scale"])
            other_shape = params["scale"].shape

        else:
            assert len(other.shape) > 0, "other must be vector/matrix"

            dep = (self.mfem_numba_coeff, other.mfem_numba_coeff)

            params = None
            func = '\n'.join(['def f(ptx, val, scale):',
                              '    return npdot(val, scale)'])
            is_other_complex = other.complex
            other_shape = other.shape

        if len(self.shape) == 2:
            # matrix * matrix or matrix * vector
            if len(other_shape) == 1:
                assert self.shape[1] == other_shape[0], "dimenstion does not match (mat*vec)" + str(
                    self.shape) + ":" + str(other_shape)
                out_shape = (self.shape[0],)
            else:
                assert self.shape[1] == other_shape[0], "dimenstion does not match (mat*mat)" + str(
                    self.shape) + ":" + str(other_shape)
                out_shape = (self.shape[0], other_shape[1])
        elif len(self.shape) == 1:
            assert len(
                other_shape) == 1, "vector.dot only supports inner product"
            assert self.shape[0] == other_shape[0],  "dimenstion does not match (vec*vec)" + str(
                self.shape) + ":" + str(other_shape)
            out_shape = tuple()
        else:
            assert False, "should not come here (.dot)"

        is_complex = self.complex or is_other_complex

        if numba_debug:
            print("(DEBUG) numba function\n", func)
        l = {}
        exec(func, globals(), l)

        if len(out_shape) == 0:
            coeff = mfem.jit.scalar(complex=is_complex,
                                    dependency=dep,
                                    interface="simple",
                                    debug=numba_debug)(l["f"])
        elif len(out_shape) == 1:
            coeff = mfem.jit.vector(complex=is_complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    shape=out_shape)(l["f"])

        elif len(out_shape) == 2:
            coeff = mfem.jit.matrix(complex=is_complex,
                                    dependency=dep,
                                    debug=numba_debug,
                                    interface="simple",
                                    shape=out_shape)(l["f"])

        else:
            assert False, "unsupported dim: dim=" + str(self.ndim)

        return NumbaCoefficient(coeff)

    def __abs__(self, other):
        raise NotImplementedError

    def __pow__(self, exponent):
        raise NotImplementedError

    def __getitem__(self, arg):
        check = self.kind == 'matrix' or self.kind == 'vector'
        assert check, "slice is valid for vector and matrix"

        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        coeff = None
        dep = (self.mfem_numba_coeff, )

        if self.kind == "vector":
            slice1 = arg

            try:
                a = slice1[0]
            except:
                slice1 = (slice1, )

            func = '\n'.join(['def f(ptx, coeff1):',
                              '    return coeff1[slice1]'])
            if numba_debug:
                print("(DEBUG) numba function\n", func)
            l = {}
            exec(func, globals(), l)

            if len(slice1) == 1:
                params = {"slice1": slice1[0]}
                coeff = mfem.jit.scalar(complex=self.complex,
                                        dependency=dep,
                                        params=params,
                                        debug=numba_debug)(l["f"])
            else:
                params = {"slice1": array(slice1, dtype=int)}
                coeff = mfem.jit.vector(complex=self.complex,
                                        dependency=dep,
                                        params=params,
                                        shape=(len(slice1),),
                                        debug=numba_debug)(l["f"])

        if self.kind == "matrix":
            slice1, slice2 = arg
            try:
                a = slice1[0]
            except:
                slice1 = (slice1, )
            try:
                a = slice2[0]
            except:
                slice2 = (slice2, )

            func1 = '\n'.join(['def f(ptx, coeff1):',
                               '    return coeff1[slice1, slice2]'])
            func2 = '\n'.join(['def f(ptx, coeff1, out):',
                               '    for ii in range(shape[0]):',
                               '        out[ii] = coeff1[slice1, slice2[ii]]'])
            func3 = '\n'.join(['def f(ptx, coeff1, out):',
                               '    for ii in range(shape[0]):',
                               '        out[ii] = coeff1[slice1[ii], slice2]'])
            func4 = '\n'.join(['def f(ptx, coeff1, out):',
                               '    for ii in range(shape[0]):',
                               '        for jj in range(shape[1]):',
                               '            out[ii, jj] = coeff1[slice1[ii], slice2[jj]]'])

            l = {}
            if len(slice1) == 1 and len(slice2) == 1:
                params = {"slice1": slice1[0], "slice2": slice2[0]}
                if numba_debug:
                    print("(DEBUG) numba function\n", func1)
                exec(func1, globals(), l)
                coeff = mfem.jit.scalar(complex=self.complex,
                                        dependency=dep,
                                        params=params,
                                        debug=numba_debug)(l["f"])
            elif len(slice1) == 1 and len(slice2) > 1:
                params = {"slice1": slice1[0],
                          "slice2": array(slice2, dtype=int)}
                if numba_debug:
                    print("(DEBUG) numba function\n", func2)

                exec(func2, globals(), l)
                coeff = mfem.jit.vector(complex=self.complex,
                                        dependency=dep,
                                        params=params,
                                        shape=(len(slice2),),
                                        interface="c++",
                                        debug=numba_debug)(l["f"])
            elif len(slice1) > 1 and len(slice2) == 1:
                params = {"slice1": array(slice1, dtype=int),
                          "slice2": slice2[0]}
                if numba_debug:
                    print("(DEBUG) numba function\n", func3)

                exec(func3, globals(), l)
                coeff = mfem.jit.vector(complex=self.complex,
                                        dependency=dep,
                                        params=params,
                                        shape=(len(slice1),),
                                        interface="c++",
                                        debug=numba_debug)(l["f"])
            else:
                params = {"slice1": array(slice1, dtype=int),
                          "slice2": array(slice2, dtype=int), }
                if numba_debug:
                    print("(DEBUG) numba function\n", func4)
                exec(func4, globals(), l)
                coeff = mfem.jit.matrix(complex=self.complex,
                                        dependency=dep,
                                        params=params,
                                        shape=(len(slice1), len(slice2)),
                                        interface="c++",
                                        debug=numba_debug)(l["f"])

        assert coeff is not None, "coeff is not build during __getitem__ in NumbaCoefficient"
        return NumbaCoefficient(coeff)

    def inv(self):
        from petram.mfem_config import numba_debug
        numba_debug = False if myid != 0 else numba_debug

        func = '\n'.join(['def f(ptx, coeff1):',
                          '    return inv(coeff1)'])
        l = {}
        if numba_debug:
            print("(DEBUG) numba function\n", func)
        exec(func, globals(), l)

        dep = (self.mfem_numba_coeff, )

        coeff = mfem.jit.matrix(complex=self.complex,
                                dependency=dep,
                                shape=self.shape,
                                interface="simple",
                                debug=numba_debug)(l["f"])
        return NumbaCoefficient(coeff)

    def adj(self):
        from petram.mfem_config import numba_debug

        func = '\n'.join(['def f(ptx, mat):',
                          '      determinant = det(mat)',
                          '      mat = inv(mat)',
                          '      return mat * determinant'])

        l = {}
        if numba_debug:
            print("(DEBUG) numba function\n", func)
        exec(func, globals(), l)

        dep = (self.mfem_numba_coeff, )

        coeff = mfem.jit.matrix(complex=self.complex,
                                dependency=dep,
                                shape=self.shape,
                                interface="simple",
                                debug=numba_debug)(l["f"])
        return NumbaCoefficient(coeff)


'''

  convert text to numba jittec coefficient

'''


def _expr_to_numba_coeff(txt, jitter, ind_vars, conj, scale, g, l,
                         return_complex, **kwargs):

    ind_vars = [xx.strip() for xx in ind_vars.split(',')]
    code = compile(txt.strip(), '<string>', 'eval')
    names = code.co_names

    dependency = []
    dep_names = []

    l = l.copy()
    for n in g:
        if isinstance(g[n], Variable):
            if n not in l:
                l[n] = g[n]
            else:
                assert False, "name confict: " + n + " is defined in local namespace"

    for n in names:
        if n in ind_vars:
            continue

        dep = None
        if n in l:
            if isinstance(l[n], Variable):
                gg = l[n]
                dep = gg.get_jitted_coefficient(ind_vars, l)
            else:
                continue
        elif n in g:
            # if isinstance(g[n], Variable):
            #    gg = g[n]
            #    dep = gg.get_jitted_coefficient(ind_vars, l)
            if isinstance(g[n], NativeCoefficientGenBase):
                from petram.phys.coefficient import call_nativegen
                c1 = call_nativegen(g[n], l, g, True, conj, scale)
                c2 = call_nativegen(g[n], l, g, False, conj, scale)
                dep = (c1, c2)
            else:
                continue
        if dep is None:
            #
            return None

        dependency.append(dep)
        dep_names.append(n)

    if jitter == mfem.jit.scalar:
        return_type = "complex128" if return_complex else "float64"
    elif jitter == mfem.jit.vector:
        return_type = "complex128[:]" if return_complex else "float64[:]"
    elif jitter == mfem.jit.matrix:
        return_type = "complex128[:,:]" if return_complex else "float64[:,:]"
    else:
        assert False, "unknown jitter: " + str(type(jitter))

    def create_func(objmode=False):
        f0 = 'def _func_(ptx, '
        for n in dep_names:
            f0 += n + ', '
        f0 += '):'

        func_txt = [f0]
        for k, xx in enumerate(ind_vars):
            func_txt.append("   " + xx + " = ptx[" + str(k) + "]")
        if objmode:
            func_txt.append("   with objmode(_out_='"+return_type+"'):")
            func_txt.append("       _out_ =" + txt)
        else:
            func_txt.append("   _out_ =" + txt)
        func_txt.append("   if isinstance(_out_, list):")
        func_txt.append("         _out_ = np.array(_out_)")
        func_txt.append("   elif isinstance(_out_, tuple):")
        func_txt.append("         _out_ = np.array(_out_)")
        if scale != 1:
            func_txt.append("   _out_ = _out_ * " + str(scale))
        if conj:
            func_txt.append("   _out_ = np.conj(_out_)")
        # func_txt.append("   print(_out_)")

        if jitter == mfem.jit.scalar:
            if return_complex:
                func_txt.append("   return np.complex128(_out_)")
            else:
                func_txt.append("   return np.float64(_out_)")
        else:
            if return_complex:
                func_txt.append("   return _out_.astype(np.complex128)")
            else:
                func_txt.append("   return _out_.astype(np.float64)")
        return func_txt

    func_txt = "\n".join(create_func())

    from petram.mfem_config import (numba_debug,
                                    get_allow_python_function_coefficient)

    numba_debug = False if myid != 0 else numba_debug

    if numba_debug:
        print("(DEBUG) wrapper function\n", func_txt)
    exec(func_txt, g, l)

    try:
        import traceback
        coeff = jitter(sdim=len(ind_vars), complex=return_complex, debug=numba_debug,
                       dependency=dependency, **kwargs)(l["_func_"])
        del l["_func_"]
    except AssertionError:
        if get_allow_python_function_coefficient() == "error":
            if myid == 0:
                traceback.print_exc()
                print("Can not JIT coefficient")
            return None

        if get_allow_python_function_coefficient() == "warn":
            if myid == 0:
                traceback.print_exc()
                print("problematic function is following")
                print(func_txt)

        if myid == 0:
            print("!!!! Failed to compile with nonpython mode. (next) Try object mode")

    except BaseException:
        if get_allow_python_function_coefficient() == "error":
            if myid == 0:
                traceback.print_exc()
                print("Can not JIT coefficient")
            return None

        if get_allow_python_function_coefficient() == "warn":
            if myid == 0:
                traceback.print_exc()
                print("problematic function is following")
                print(func_txt)

        if myid == 0:
            print("!!!! Failed to compile with nonpython mode. (next) Try object mode")

    else:
        return NumbaCoefficient(coeff)

    # for the moment suppress objmode creation
    # return None

    func_txt = "\n".join(create_func(True))

    from petram.mfem_config import (numba_debug,
                                    get_allow_python_function_coefficient)

    numba_debug = False if myid != 0 else numba_debug

    if numba_debug:
        print("(DEBUG) wrapper function\n", func_txt)

    exec(func_txt, g, l)

    try:
        from numba import objmode
        coeff = jitter(sdim=len(ind_vars), complex=return_complex, debug=numba_debug,
                       dependency=dependency, params={"objmode": objmode},
                       **kwargs)(l["_func_"])
        del l["_func_"]
    except AssertionError:
        import traceback
        traceback.print_exc()

        print("Can not JIT coefficient. No possible recoverly. ")
        return None
    except BaseException:
        import traceback
        traceback.print_exc()
        print("Can not JIT coefficient. No possible recovery.")
        return None

    return NumbaCoefficient(coeff)


def expr_to_numba_coeff(exprs, jitter, ind_vars, conj, scale, g, l, return_complex,
                        **kwargs):
    '''
    # generate a wrapper for multiple inputs
    def _func_(ptx, p1, p2, out)
        deps = [p1, p2]
        count = 0
        depcount = 0
        for i0 in range(shape[0]):
            for i1 in range(shape[1]):
                if isconst[count]:
                   value = consts[count]
                else:
                   value = deps[depcount]
                   depcount = depcount + 1
                out[i0:i1] = value
                count = count +1
    '''

    consts = [-1]*len(exprs)
    isconst = zeros(len(exprs))
    deps = []
    dep_names = []

    from petram.mfem_config import get_allow_python_function_coefficient
    if get_allow_python_function_coefficient() == "always use Python coeff.":
        return None

    if len(exprs) > 1:
        jitter2 = mfem.jit.scalar
        assert "shape" in kwargs, "multile expression array, but shape is not given"
        shape = kwargs.pop("shape")
    else:
        jitter2 = jitter

    nbcs = []
    for k, ee in enumerate(exprs):
        if isinstance(ee, str):
            nbc = _expr_to_numba_coeff(
                ee, jitter2, ind_vars, conj, scale, g, l, return_complex, **kwargs)
            if nbc is None:
                return None
            deps.append(nbc.mfem_numba_coeff)
            dep_names.append("p"+str(k))
            nbcs.append(nbc)
        else:
            isconst[k] = 1
            if scale != 1:
                ee = ee * scale
            if conj:
                ee = npconj(ee)
            consts[k] = ee
    if len(exprs) == 1:
        return nbc

    consts = array(consts)
    ind_vars = [xx.strip() for xx in ind_vars.split(',')]

    f0 = 'def _func_(ptx, '
    for n in dep_names:
        f0 += n + ', '
    f0 += 'out):'

    func_txt = [f0]
    func_txt.append("    count = 0")
    func_txt.append("    depcount = 0")
    func_txt.append("    deps = [" + ",".join(dep_names) + "]")

    idx_text = ""
    for k, s in enumerate(shape):
        func_txt.append("    " + " "*k + "for i" + str(k) +
                        " in range(" + str(s) + "):")
        idx_text = idx_text + "i"+str(k)+","

    func_txt.append("     " + " "*len(shape) + "if isconst[count] == 1:")
    func_txt.append("     " + " "*len(shape) + "    value = consts[count]")
    func_txt.append("     " + " "*len(shape) + "else:")
    func_txt.append("     " + " "*len(shape) + "    value = deps[depcount]")
    func_txt.append("     " + " "*len(shape) + "    depcount = depcount + 1")
    func_txt.append("     " + " "*len(shape) + "out["+idx_text + "]=value")
    func_txt.append("     " + " "*len(shape) + "count = count + 1")

    func_txt = "\n".join(func_txt)

    g = globals()
    l = {}

    from petram.mfem_config import numba_debug
    numba_debug = False if myid != 0 else numba_debug

    if numba_debug:
        print("(DEBUG) wrapper function\n", func_txt)
    exec(func_txt, g, l)

    params = {}
    params["isconst"] = isconst
    params["consts"] = consts

    coeff = jitter(sdim=len(ind_vars),
                   complex=return_complex,
                   debug=numba_debug,
                   shape=shape,
                   dependency=deps,
                   interface="c++",
                   params=params,
                   **kwargs)(l["_func_"])

    ret = NumbaCoefficient(coeff)
    del l["_func_"]
    return ret


'''
convert function to numba coefficient
'''


def func_to_numba_coeff_scalar(func, complex=False,
                               params=None, dependency=None):
    from petram.mfem_config import numba_debug
    numba_debug = False if myid != 0 else numba_debug

    if dependency is None:
        dependency = tuple()

    dependency = [(x.mfem_numba_coeff if isinstance(x, NumbaCoefficient) else x)
                  for x in dependency]

    jitter = mfem.jit.scalar(complex=complex, params=params,
                             debug=numba_debug, dependency=dependency)

    mfem_coeff1 = jitter(func)
    return NumbaCoefficient(mfem_coeff1)


def func_to_numba_coeff_vector(func, shape=(0,), complex=False,
                               params=None, dependency=None):
    from petram.mfem_config import numba_debug
    numba_debug = False if myid != 0 else numba_debug

    if dependency is None:
        dependency = tuple()

    dependency = [(x.mfem_numba_coeff if isinstance(x, NumbaCoefficient) else x)
                  for x in dependency]

    jitter = mfem.jit.vector(shape=shape, complex=complex, params=params,
                             debug=numba_debug, dependency=dependency)

    mfem_coeff1 = jitter(func)
    return NumbaCoefficient(mfem_coeff1)


def func_to_numba_coeff_matrix(func, shape=(0, 0), complex=False,
                               params=None, dependency=None):

    from petram.mfem_config import numba_debug
    numba_debug = False if myid != 0 else numba_debug

    if dependency is None:
        dependency = tuple()

    dependency = [(x.mfem_numba_coeff if isinstance(x, NumbaCoefficient) else x)
                  for x in dependency]

    jitter = mfem.jit.matrix(shape=shape, complex=complex, params=params,
                             debug=numba_debug, dependency=dependency)

    mfem_coeff1 = jitter(func)
    return NumbaCoefficient(mfem_coeff1)



from __future__ import print_function

'''
   matrix_file 
   
   a group of helper routine to read and write matrix/vector to file.


'''
import numpy as np
import os
import six


def read_matvec(file, all=False, verbose=False, complex=False, skip=0):
    '''
    read matrix/vector file.  
    If all is on, read all files with same basename ('matrix.0000, matrix.0001...')
    '''
    if not all:
        files = [file]
    else:
        dir = os.path.dirname(file)
        base = os.path.basename(file)
        files = []
        for x in os.listdir(dir):
            if x.find(base) != -1:
                files.append(x)
        files = sorted(files)
        files = [os.path.join(dir, f) for f in files]
        if verbose:
            six.print_(files)

    if len(files) == 0:
        return

    ret = []
    for file in files:
        fid = open(file, "r")
        xx = [x.strip().split() for x in fid.readlines()]
        xx = xx[skip:]
        if complex:
            xxx = [[np.complex(x) for x in y] for y in xx]
        else:
            xxx = [[np.float(x) for x in y] for y in xx]
        fid.close()
        ret.append(np.array(xxx))
    return np.vstack(ret)


def write_matrix(file, m):
    from petram.mfem_config import use_parallel
    if use_parallel:
        from mpi4py import MPI
        num_proc = MPI.COMM_WORLD.size
        myid = MPI.COMM_WORLD.rank
        smyid = '.'+'{:0>6d}'.format(myid)
    else:
        smyid = ''
    if hasattr(m, 'save_data'):
        m.save_data(file + smyid)
    else:
        raise NotImplemented("write matrix not implemented for" + m.__repr__())


def write_vector(file, bb):
    from petram.mfem_config import use_parallel
    if use_parallel:
        from mpi4py import MPI
        num_proc = MPI.COMM_WORLD.size
        myid = MPI.COMM_WORLD.rank
        smyid = '.'+'{:0>6d}'.format(myid)
    else:
        smyid = ''

    if hasattr(bb, "SaveToFile"):   # GridFunction
        bb.SaveToFile(file+smyid, 8)
    else:
        fid = open(file+smyid, "w")
        if np.iscomplexobj(bb):
            for k, x in enumerate(bb):
                fid.write(str(k) + ' ' + "{0:.8g}".format(x.real) + ' ' +
                          "{0:.8g}".format(x.imag) + '\n')
        else:
            for k, x in enumerate(bb):
                fid.write(str(k) + ' ' + "{0:.8g}".format(x) + '\n')
        fid.close()


def write_coo_matrix(file, A):
    from petram.mfem_config import use_parallel
    if use_parallel:
        from mpi4py import MPI
        num_proc = MPI.COMM_WORLD.size
        myid = MPI.COMM_WORLD.rank
        smyid = '.'+'{:0>6d}'.format(myid)
    else:
        smyid = ''

    if (A.dtype == 'complex'):
        is_complex = True
    else:
        is_complex = False

    fid = open(file+smyid, 'w')
    rc = np.vstack((A.row, A.col)).transpose()
    tmp = sorted([(k, tuple(x)) for k, x in enumerate(rc)], key=lambda x: x[1])
    idx = np.array([x[0] for x in tmp])
    if len(idx) == 0:
        fid.close()
        return

    row = A.row[idx]
    col = A.col[idx]
    data = A.data[idx]

    if is_complex:
        txt = [' '.join([str(int(r)), str(int(c)), "{0:.8g}".format(a.real),
                         "{0:.8g}".format(a.imag)]) for r, c, a in zip(row, col, data)]
        fid.write('\n'.join(txt) + "\n")
    else:
        txt = [' '.join([str(int(r)), str(int(c)), "{0:.8g}".format(a.real),
                         "{0:.8g}".format(a.imag)]) for r, c, a in zip(row, col, data)]
        fid.write('\n'.join(txt) + "\n")
    fid.close()



'''

   Weakform : interface to use MFEM integrator

'''
import os
import numpy as np

from petram.phys.phys_model import Phys
from petram.phys.phys_model  import PhysCoefficient
from petram.phys.phys_model  import VectorPhysCoefficient
from petram.phys.phys_model  import MatrixPhysCoefficient
from petram.model import Domain, Bdry, Edge, Point, Pair

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('WeakForm')

from petram.mfem_config import use_parallel
if use_parallel:
   import mfem.par as mfem
else:
   import mfem.ser as mfem
   
from petram.phys.vtable import VtableElement, Vtable   


data =  (('prj_name', VtableElement('prj_name', type='string',
                                     guilabel = 'name',
                                     default = '',
                                     tip = "name of projected value")),
         ('src_trans', VtableElement('stc_trans', type='string',
                                     guilabel = 'transform(src)',
                                     default = '',
                                     tip = "src coordinate transformation")),
         ('dst_trans', VtableElement('dst_trans', type='string',
                                     guilabel = 'transform(dst)',
                                     default = '',
                                     tip = "src coordinate transformation")),
         ('prj_factor', VtableElement('prj_factor', type='any',
                                      guilabel = 'factor',
                                      default = '',
                                      tip = "src coordinate transformation")),
         ('mapping_tol', VtableElement('mapping_tol', type='any',
                                       guilabel = 'mapping tolerance',
                                       default = '',
                                       tip = "tolerance of DoF mapping")))

class DoFProjection(Pair, Phys):
    vt  = Vtable(data)
    def __init__(self, **kwargs):    
        super(DoFProjection, self).__init__(**kwargs)
        Phys.__init__(self)
    
    def attribute_set(self, v):
        v = super(DoFProjection, self).attribute_set(v)
        v['prj_fes_idx'] = 0
        return v
    
    def panel1_param(self):
        import wx
        names = self.get_root_phys().dep_vars
        ll = [["FESpace", "S", 4,
               {"style":wx.CB_READONLY, "choices": names}]]
        param = self.vt.panel_param(self)
        ll.extend(param)
        return ll

    def import_panel1_value(self, v):
        idx = self.get_root_phys().dep_vars.index(v[0])
        self.prj_fes_idx = idx
        return self.vt.import_panel_value(self, v[1:])
    
    def get_panel1_value(self):
        val = [self.get_root_phys().dep_vars[self.prj_fes_idx]]
        val.extend(self.vt.get_panel_value(self))
        return val
    
    def panel1_tip(self):
        pass

    def assemble_matrix(self):
        raise NotImplementedError("subclass must implement this")
    
class BdrProjection(DoFProjection, Bdry):
    pass
class DomainProjection(DoFProjection, Domain):
    pass




'''

   coefficient generation funcitons

'''
from petram.debug import handle_allow_python_function_coefficient
import numpy as np

from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem


from petram.phys.phys_model import (PhysCoefficient,
                                    VectorPhysCoefficient,
                                    MatrixPhysCoefficient,
                                    PhysConstant,
                                    PhysVectorConstant,
                                    PhysMatrixConstant,
                                    Coefficient_Evaluator)

from petram.helper.variables import (Variable,
                                     NativeCoefficientGenBase)

from petram.phys.pycomplex_coefficient import (CC_Matrix,
                                               CC_Vector,
                                               CC_Scalar,
                                               PyComplexConstant,
                                               PyComplexVectorConstant,
                                               PyComplexMatrixConstant,
                                               complex_coefficient_from_real_and_imag)

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('Coefficient')


def call_nativegen(v, l, g, real, conj, scale):
    vv = v(l, g)
    if real:
        coeff = vv[0]
        if scale != 1.0 and coeff is not None:
            coeff = v.scale_coeff(coeff, scale)

        return coeff
    else:
        if conj:
            assert False, "conj is not supported for NativeCoefficient"
        else:
            coeff = vv[1]
            if scale != 1.0 and coeff is not None:
                coeff = v.scale_coeff(coeff, scale)

            return coeff


def MCoeff(dim, exprs, ind_vars, l, g, return_complex=False,
           return_mfem_constant=False, **kwargs):
    if isinstance(exprs, str):
        exprs = [exprs]
    if isinstance(exprs, NativeCoefficientGenBase):
        exprs = [exprs]

    class MCoeff_Base(object):
        def __init__(self, conj=False, scale=1.0):
            self.conj = conj
            self.scale = scale

        def proc_value(self, val):
            val = val * self.scale
            if self.conj:
                val = np.conj(val)
            return val

    class MCoeff(MatrixPhysCoefficient, MCoeff_Base):
        def __init__(self, sdim, exprs, ind_vars, l, g,
                     scale=1.0, conj=False, **kwargs):
            MCoeff_Base.__init__(self, conj=conj, scale=scale)
            MatrixPhysCoefficient.__init__(
                self, sdim, exprs, ind_vars, l, g, **kwargs)

        def EvalValue(self, x):
            val = super(MCoeff, self).EvalValue(x)
            val = self.proc_value(val)

            if np.iscomplexobj(val):
                if self.real:
                    return val.real
                else:
                    return val.imag
            elif not self.real:
                return val * 0.0
            else:
                return val

    class MCoeffCC(Coefficient_Evaluator, MCoeff_Base, CC_Matrix):
        def __init__(self, dim, exprs, ind_vars, l,
                     g, conj=False, scale=1.0, **kwargs):
            MCoeff_Base.__init__(self, conj=conj, scale=scale)
            # real is not used...
            Coefficient_Evaluator.__init__(
                self, exprs, ind_vars, l, g, real=True)
            CC_Matrix.__init__(self, dim, dim)

        def eval(self, T, ip):
            for n, v in self.variables:
                v.set_point(T, ip, self.g, self.l)
            x = T.Transform(ip)
            val = Coefficient_Evaluator.EvalValue(self, x)
            val = val.reshape(self.height, self.width)
            return self.proc_value(val)

    conj = kwargs.get('conj', False)
    real = kwargs.get('real', True)
    scale = kwargs.get('scale', 1.0)

    if any([isinstance(ee, str) for ee in exprs]):
        # if it is one liner array expression. try mfem.jit
        from petram.phys.numba_coefficient import expr_to_numba_coeff

        coeff = expr_to_numba_coeff(exprs, mfem.jit.matrix,
                                    ind_vars, conj, scale, g, l,
                                    return_complex, shape=(dim, dim))
        if coeff is None:
            msg = "JIT is not possbile. Continuing with Python mode"
            handle_allow_python_function_coefficient(msg)
        else:
            if return_complex:
                return coeff
            else:
                if real:
                    return coeff.real
                else:
                    return coeff.imag

        if return_complex:
            return MCoeffCC(dim, exprs, ind_vars, l, g, **kwargs)
        else:
            return MCoeff(dim, exprs, ind_vars, l, g, **kwargs)
    else:
        e = exprs

        if isinstance(e[0], NativeCoefficientGenBase):
            if return_complex:
                c1 = call_nativegen(e[0], l, g, True, conj, scale)
                c2 = call_nativegen(e[0], l, g, False, conj, scale)
                return complex_coefficient_from_real_and_imag(c1, c2)
            else:
                return call_nativegen(e[0], l, g, real, conj, scale)

        e = np.array(e, copy=False).reshape(dim, dim)
        e = e * scale
        if conj:
            e = np.conj(e)

        if return_complex:
            assert not return_mfem_constant, "return_complex and return_mfem_constant can not be used togeter"
            e = e.astype(complex)
            return PyComplexMatrixConstant(e)
        else:
            if np.iscomplexobj(e):
                if real:
                    e = e.real
                else:
                    e = e.imag
            elif not real:
                e = np.array(e * 0.0, dtype=float, copy=False)
            else:
                e = np.array(e, dtype=float, copy=False)

            if return_mfem_constant:
                return mfem.MatrixConstantCoefficient(e)
            else:
                return PhysMatrixConstant(e)


def DCoeff(dim, exprs, ind_vars, l, g,
           return_mfem_constant=False, **kwargs):
    if isinstance(exprs, str):
        exprs = [exprs]
    if isinstance(exprs, NativeCoefficientGenBase):
        exprs = [exprs]

    class DCoeff(MatrixPhysCoefficient):
        def __init__(self, *args, **kwargs):
            self.conj = kwargs.pop('conj', False)
            self.scale = kwargs.pop('scale', 1.0)
            self.space_dim = args[0]
            super(DCoeff, self).__init__(*args, **kwargs)

        def EvalValue(self, x):
            from petram.phys.phys_model import Coefficient_Evaluator
            val = Coefficient_Evaluator.EvalValue(self, x)
            val = np.diag(val)
            val = val * self.scale
            if self.conj:
                val = np.conj(val)

            if np.iscomplexobj(val):
                if self.real:
                    return val.real
                else:
                    return val.imag
            elif not self.real:
                return val * 0.0
            else:
                return val

    conj = kwargs.get('conj', False)
    real = kwargs.get('real', True)
    scale = kwargs.get('scale', 1.0)

    #print("matrix exprs", exprs)

    if any([isinstance(ee, str) for ee in exprs]):
        from petram.mfem_config import allow_python_function_coefficient
        if allow_python_function_coefficient != "allow":
            print("JIT is not supported in DCoeff")
            assert False, "can not jit coefficient"
        return DCoeff(dim, exprs, ind_vars, l, g, **kwargs)
    else:
        e = exprs

        if isinstance(e[0], NativeCoefficientGenBase):
            return call_nativegen(e[0], l, g, real, conj, scale)

        e = e * scale
        e = np.diag(e)
        if np.iscomplexobj(e):
            if conj:
                e = np.conj(e)
            if real:
                e = e.real
            else:
                e = e.imag
        elif not real:
            e = np.array(e * 0.0, dtype=float, copy=False)
        else:
            e = np.array(e, dtype=float, copy=False)

        if return_mfem_constant:
            return mfem.MatrixConstantCoefficient(e)
        else:
            return PhysMatrixConstant(e)


def VCoeff(dim, exprs, ind_vars, l, g, return_complex=False,
           return_mfem_constant=False, **kwargs):
    if isinstance(exprs, str):
        exprs = [exprs]
    if isinstance(exprs, NativeCoefficientGenBase):
        exprs = [exprs]

    class Vcoeff_Base(object):
        def __init__(self, conj=False, scale=1.0):
            self.conj = conj
            self.scale = scale

        def proc_value(self, val):
            val = val * self.scale
            if self.conj:
                val = np.conj(val)
            return val

    class VCoeff(VectorPhysCoefficient, Vcoeff_Base):
        def __init__(self, dim, exprs, ind_vars, l, g,
                     conj=False, scale=1.0, **kwargs):
            Vcoeff_Base.__init__(self, conj=conj, scale=scale)
            VectorPhysCoefficient.__init__(
                self, dim, exprs, ind_vars, l, g, **kwargs)

        def EvalValue(self, x):
            val = super(VCoeff, self).EvalValue(x)
            val = self.proc_value(val)

            if np.iscomplexobj(val):
                if self.real:
                    return val.real
                else:
                    return val.imag
            elif not self.real:
                return val * 0.0
            else:
                return val

    class VCoeffCC(Coefficient_Evaluator, Vcoeff_Base, CC_Vector):
        def __init__(self, dim, exprs, ind_vars, l,
                     g, conj=False, scale=1.0, **kwargs):
            Vcoeff_Base.__init__(self, conj=conj, scale=scale)
            # real is not used...
            Coefficient_Evaluator.__init__(
                self, exprs, ind_vars, l, g, real=True)
            CC_Vector.__init__(self, dim)

        def eval(self, T, ip):
            for n, v in self.variables:
                v.set_point(T, ip, self.g, self.l)
            x = T.Transform(ip)
            val = Coefficient_Evaluator.EvalValue(self, x)
            return self.proc_value(val)

    conj = kwargs.get('conj', False)
    real = kwargs.get('real', True)
    scale = kwargs.get('scale', 1.0)

    #print("vector exprs", exprs)

    if any([isinstance(ee, str) for ee in exprs]):
        # if it is one liner array expression. try mfem.jit
        from petram.phys.numba_coefficient import expr_to_numba_coeff
        coeff = expr_to_numba_coeff(exprs, mfem.jit.vector,
                                    ind_vars, conj, scale, g, l,
                                    return_complex, shape=(dim, ))
        if coeff is None:
            msg = "JIT is not possbile. Continuing with Python mode"
            handle_allow_python_function_coefficient(msg)

        else:
            if return_complex:
                return coeff
            else:
                if real:
                    return coeff.real
                else:
                    return coeff.imag

        if return_complex:
            return VCoeffCC(dim, exprs, ind_vars, l, g, **kwargs)
        else:
            return VCoeff(dim, exprs, ind_vars, l, g, **kwargs)

    else:
        e = exprs

        if isinstance(e[0], NativeCoefficientGenBase):
            if return_complex:
                c1 = call_nativegen(e[0], l, g, True, conj, scale)
                c2 = call_nativegen(e[0], l, g, False, conj, scale)
                return complex_coefficient_from_real_and_imag(c1, c2)
            else:
                return call_nativegen(e[0], l, g, real, conj, scale)

        e = np.array(e, copy=False)
        e = e * scale

        if return_complex:
            assert not return_mfem_constant, "return_complex and return_mfem_constant can not be used togeter"
            e = e.astype(complex)
            return PyComplexVectorConstant(e)
        else:
            if np.iscomplexobj(e):
                if conj:
                    e = np.conj(e)
                if real:
                    e = e.real
                else:
                    e = e.imag
            elif not real:
                e = np.array(e * 0.0, dtype=float, copy=False)
            else:
                e = np.array(e, dtype=float, copy=False)
            if return_mfem_constant:
                return mfem.VectorConstantCoefficient(e)
            else:
                return PhysVectorConstant(e)


def SCoeff(exprs, ind_vars, l, g, return_complex=False,
           return_mfem_constant=False, **kwargs):
    if isinstance(exprs, str):
        exprs = [exprs]
    if isinstance(exprs, NativeCoefficientGenBase):
        exprs = [exprs]

    class Scoeff_Base(object):
        def __init__(self, component=None, conj=False, scale=1.0):
            self.component = component
            self.conj = conj
            self.scale = scale

        def proc_value(self, val):
            if self.component is None:
                if self.conj:
                    val = np.conj(val)
                v = val
            else:
                if len(val.shape) == 0:
                    val = [val]
                if self.conj:
                    v = np.conj(val)[self.component]
                else:
                    v = val[self.component]
            v = v * self.scale
            return v

    class SCoeff(PhysCoefficient, Scoeff_Base):
        def __init__(self, exprs, ind_vars, l, g, component=None,
                     conj=False, scale=1.0, **kwargs):
            Scoeff_Base.__init__(
                self,
                component=component,
                conj=conj,
                scale=scale)
            super(SCoeff, self).__init__(exprs, ind_vars, l, g, **kwargs)

        def EvalValue(self, x):
            val = super(SCoeff, self).EvalValue(x)
            v = self.proc_value(val)
            if np.iscomplexobj(v):
                if self.real:
                    return v.real
                else:
                    return v.imag
            elif not self.real:
                return 0.0
            else:
                return v

    class SCoeffCC(Coefficient_Evaluator, Scoeff_Base, CC_Scalar):
        def __init__(self, exprs, ind_vars, l, g,
                     component=None, conj=False, scale=1.0, **kwargs):
            Scoeff_Base.__init__(
                self,
                component=component,
                conj=conj,
                scale=scale)
            # real is not used...
            Coefficient_Evaluator.__init__(
                self, exprs, ind_vars, l, g, real=True)
            CC_Scalar.__init__(self)

        def eval(self, T, ip):
            for n, v in self.variables:
                v.set_point(T, ip, self.g, self.l)
            x = T.Transform(ip)
            val = Coefficient_Evaluator.EvalValue(self, x)
            if len(self.co) == 1 and len(val) == 1:
                val = val[0]
            return self.proc_value(val)

    component = kwargs.get('component', None)
    conj = kwargs.get('conj', False)
    real = kwargs.get('real', True)
    scale = kwargs.get('scale', 1.0)

    if any([isinstance(ee, str) for ee in exprs]):
        if len(exprs) == 1:
            # if it is one liner array expression. try mfem.jit
            from petram.phys.numba_coefficient import expr_to_numba_coeff
            coeff = expr_to_numba_coeff(exprs, mfem.jit.scalar,
                                        ind_vars, conj, scale, g, l, return_complex)
            if coeff is None:
                msg = "JIT is not possbile. Continuing with Python mode"
                handle_allow_python_function_coefficient(msg)
            else:
                if return_complex:
                    return coeff
                else:
                    if real:
                        return coeff.real
                    else:
                        return coeff.imag
        else:
            # should not come here
            assert False, "Scalar coefficient can not use mutliple expressions"

        if return_complex:
            return SCoeffCC(exprs, ind_vars, l, g, **kwargs)
        else:
            return SCoeff(exprs, ind_vars, l, g, **kwargs)
    else:
        # conj is ignored..(this doesn't no meaning...)
        # print("exprs",exprs)
        if component is None:
            v = exprs[0]  # exprs[0]
        else:
            # weakform10 didn't work with-> exprs[0][component]
            v = exprs[component]

        if isinstance(v, NativeCoefficientGenBase):
            if return_complex:
                c1 = call_nativegen(v, l, g, True, conj, scale)
                c2 = call_nativegen(v, l, g, False, conj, scale)
                return complex_coefficient_from_real_and_imag(c1, c2)
            else:
                return call_nativegen(v, l, g, real, conj, scale)

        v = v * scale

        if return_complex:
            v = complex(v)
            if conj:
                v = np.conj(v)
            assert not return_mfem_constant, "return_complex and return_mfem_constant can not be used togeter"
            return PyComplexConstant(v)
        else:
            if np.iscomplexobj(v):
                if conj:
                    v = np.conj(v)
                if real:
                    v = v.real
                else:
                    v = v.imag
            elif not real:
                v = 0.0
            else:
                pass
            v = float(v)
            if return_mfem_constant:
                return mfem.ConstantCoefficient(v)
            else:
                return PhysConstant(v)


def sum_coefficient(c_arr):
    '''
    return sum_coefficient made from list of coefficient
    '''
    if len(c_arr) == 0:
        return None
    if len(c_arr) == 1:
        return c_arr[0]
    kind = ''
    for c in c_arr:
        if isinstance(c, mfem.Coefficient):
            if kind != '' and kind != 's':
                assert False, "can not mix diffenrnt kind of coefficient"
            kind = 's'
        if isinstance(c, mfem.VectorCoefficient):
            if kind != '' and kind != 'v':
                assert False, "can not mix diffenrnt kind of coefficient"
            kind = 'v'
        if isinstance(c, mfem.MatrixCoefficient):
            if kind != '' and kind != 'm':
                assert False, "can not mix diffenrnt kind of coefficient"
            kind = 'm'

    if kind == 's':
        c = mfem.SumCoefficient(c_arr[0], c_arr[1])
        for cc in c_arr[2:]:
            c = mfem.SumCoefficient(c, cc)
    elif kind == 'v':
        c = mfem.VectorSumCoefficient(c_arr[0], c_arr[1])
        for cc in c_arr[2:]:
            c = mfem.VectorSumCoefficient(c, cc)
    elif kind == 'm':
        c = mfem.MatrixSumCoefficient(c_arr[0], c_arr[1])
        for cc in c_arr[2:]:
            c = mfem.MatrixSumCoefficient(c, cc)

    return c



'''
  DotDict 
    Dictionary container, which allows for dictionary look up by dot

'''
class DotDict(object):
    def __init__(self, d):
        self._d = d

    def __getattr__(self, attr):
      # only attributes not starting with "_" are organinzed
      # in the tree
        if not attr.startswith("_"):
            if attr in self._d:
                return self._d[attr]
            raise AttributeError('Attribute member not found :'+attr+':'+str(self))
            # return self._d.setdefault(attr, TreeDict())
        raise AttributeError('Attribute name should not start from "_"')
        



import numpy as np
import scipy
import six
from collections import defaultdict
import weakref
from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD


from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

from petram.sol.evaluator_agent import EvaluatorAgent
from petram.sol.bdr_nodal_evaluator import process_iverts2nodals
from petram.sol.bdr_nodal_evaluator import eval_at_nodals, get_emesh_idx

class EdgeNodalEvaluator(EvaluatorAgent):
    def __init__(self, attrs, plane = None):
        '''
           attrs = [[1,2,3], ax, ay, az, c]

           cut-plane is defined as
           ax * x + ay * y + ax * z + c = 0
        '''
        super(EdgeNodalEvaluator, self).__init__()
        self.attrs = attrs
        
    def preprocess_geometry(self, attrs, emesh_idx=0):
        self.vertices = None

        self.knowns = WKD()
        mesh = self.mesh()[emesh_idx]
        self.iverts = []
        self.attrs = attrs

        if attrs[0] == 'all':
            eattrs = 'all'
        else:
            eattrs = attrs


        if mesh.Dimension() == 3:
            '''
            from petram.mesh.find_edges import find_edges
            edges, bb_edges = find_edges(mesh)
            bb_bdrs = bb_edges.keys()
            iverts = []
            for bb_bdr in bb_bdrs:
                if eattrs != 'all':
                    check = [sorted(tuple(eattr)) ==  sorted(bb_bdr) for eattr in eattrs]
                    if not any(check): continue
                iedges = bb_edges[bb_bdr]
                iverts.extend([mesh.GetEdgeVertices(ie) for ie in iedges])
            print iverts
            '''
            from petram.mesh.mesh_utils import get_extended_connectivity
            if not hasattr(mesh, 'extended_connectivity'):
               get_extended_connectivity(mesh)
            
            l2e = mesh.extended_connectivity['line2edge']
            keys = l2e.keys() if eattrs == 'all' else list(np.atleast_1d(eattrs).flatten())
            iedges = list(set(sum([l2e[k] for k in keys if k in l2e],[])))
            iverts = [mesh.GetEdgeVertices(ie) for ie in iedges]
            self.ibeles = None # can not use boundary variable in this evaulator            

        elif mesh.Dimension() == 2:
            kbdr = mesh.GetBdrAttributeArray()
            if eattrs == 'all': eattrs = np.unique(kbdr)
            iverts = []
            #d = defaultdict(list)
            for i in range(mesh.GetNBE()):
                attr = mesh.GetBdrAttribute(i)
                if attr in eattrs:
                    iverts.append(list(mesh.GetBdrElement(i).GetVerticesArray()))

            x = np.unique([mesh.GetBdrArray(e) for e in eattrs]).astype(int, copy=False)
            self.ibeles = x
                    #d[attr].extend(mesh.GetBdrElement(i).GetVerticesArray())
        elif mesh.Dimension() == 1:
            kbdr = mesh.GetAttributeArray()
            if eattrs == 'all': eattrs = np.unique(kbdr)
            iverts = []
            #d = defaultdict(list)
            for i in range(mesh.GetNE()):
                attr = mesh.GetAttribute(i)
                if attr in eattrs:
                    iverts.append(list(mesh.GetElement(i).GetVerticesArray()))
                    #d[attr].extend(mesh.GetBdrElement(i).GetVerticesArray())
            x = np.unique([mesh.GetDomainArray(e) for e in eattrs]).astype(int, copy=False)
            self.ibeles = x

        else:
            assert False, "Unsupported dim"

        self.emesh_idx = emesh_idx
        
        if len(iverts) == 0: return
      
        iverts = np.stack(iverts)
        self.iverts = iverts
        if len(self.iverts) == 0: return

        data = process_iverts2nodals(mesh, iverts)
        for k in six.iterkeys(data):
            setattr(self, k, data[k])
        
            
    def eval(self, expr, solvars, phys, **kwargs):
        exprs =  kwargs.pop("exprs", [expr])
        emesh_idx = get_emesh_idx(self, exprs, solvars, phys)
        
        #print("emesh_idx", emesh_idx)
        if len(emesh_idx) > 1:
            assert False, "expression involves multiple mesh (emesh length != 1)"
        #if len(emesh_idx) < 1:
        #    assert False, "expression is not defined on any mesh"
        #(this could happen when expression is pure geometryical like "x+y")                
            
        if len(emesh_idx) == 1:
            if self.emesh_idx != emesh_idx[0]:
                 self.preprocess_geometry(self.attrs, emesh_idx=emesh_idx[0])
                 
        val = eval_at_nodals(self, expr, solvars, phys,
                             edge_evaluator=True)
        if val is None: return None, None, None

#        return self.locs[self.iverts_inv], val[self.iverts_inv, ...]

        refine = kwargs.pop('refine', 1)
        
        if refine == 1:
            return self.locs, val, self.iverts_inv
        else:
            from petram.sol.nodal_refinement import refine_edge_data
            try:
                mesh = self.mesh()[self.emesh_idx]
                if mesh.Dimension() == 3:
                    assert False, "edge refinement is not supported for dim=3"
                ptx, data, ridx = refine_edge_data(mesh,
                                                  self.ibeles,
                                                  val, self.iverts_inv,
                                                  refine)
            except:
                import traceback
                traceback.print_exc()
            return ptx, data, ridx



         
